#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <dune/common/fvector.hh>

#include <dune/tectonic/polyhedrondistance.hh>

int main() {
  using LocalVector = Dune::FieldVector<double, 2>;

  auto const test =
      [&](ConvexPolyhedron<LocalVector> const &p1,
          ConvexPolyhedron<LocalVector> const &p2, double analyticalDistance) {
        LocalVector target;
        {
          double const error =
              std::abs(analyticalDistance - distance(p1, p2, 1e-12));
          std::cout << "error: " << error << std::endl;
          assert(error < 1e-12);
        }
      };
  {
    /*
     * Calculate the distance between two triangles, where it is attained at a
     * face-vertex pair
     *
     *    O
     *    |\
     *    | \
     * O  O--O
     * |\
     * | \
     * O--O
     */
    double const analyticalDistance = std::sqrt(2.0);
    LocalVector tmp;

    ConvexPolyhedron<LocalVector> bs1;
    bs1.vertices.resize(3);
    tmp[0] = 0;
    tmp[1] = 0;
    bs1.vertices[0] = tmp;
    tmp[0] = 0;
    tmp[1] = 2;
    bs1.vertices[1] = tmp;
    tmp[0] = 2;
    tmp[1] = 0;
    bs1.vertices[2] = tmp;

    ConvexPolyhedron<LocalVector> bs2;
    bs2.vertices.resize(3);
    tmp[0] = 2;
    tmp[1] = 2;
    bs2.vertices[0] = tmp;
    tmp[0] = 2;
    tmp[1] = 4;
    bs2.vertices[1] = tmp;
    tmp[0] = 4;
    tmp[1] = 2;
    bs2.vertices[2] = tmp;

    test(bs1, bs2, analyticalDistance);
  }

  {
    /*
     * Calculate the distance between two triangles, where it is
     * attained in a face-face pair
     *
     *    O--O
     *     \ |
     *      \|
     * O     O
     * |\
     * | \
     * O--O
     */
    double const analyticalDistance = 2.0 * std::sqrt(2.0);
    LocalVector tmp;

    ConvexPolyhedron<LocalVector> bs1;
    bs1.vertices.resize(3);
    tmp[0] = 0;
    tmp[1] = 0;
    bs1.vertices[0] = tmp;
    tmp[0] = 0;
    tmp[1] = 2;
    bs1.vertices[1] = tmp;
    tmp[0] = 2;
    tmp[1] = 0;
    bs1.vertices[2] = tmp;

    ConvexPolyhedron<LocalVector> bs2;
    bs2.vertices.resize(3);
    tmp[0] = 4;
    tmp[1] = 4;
    bs2.vertices[0] = tmp;
    tmp[0] = 2;
    tmp[1] = 4;
    bs2.vertices[1] = tmp;
    tmp[0] = 4;
    tmp[1] = 2;
    bs2.vertices[2] = tmp;

    test(bs1, bs2, analyticalDistance);
  }

  {
    /*
     * Calculate the distance between two triangles, where it is
     * attained in a vertex-vertex pair
     *
     *       O
     *       |\
     *       | \
     * O--O  O--O
     *  \ |
     *   \|
     *    O
     */
    double analyticalDistance = 2.0;
    LocalVector tmp;

    ConvexPolyhedron<LocalVector> bs1;
    bs1.vertices.resize(3);
    tmp[0] = 2;
    tmp[1] = 2;
    bs1.vertices[0] = tmp;
    tmp[0] = 0;
    tmp[1] = 2;
    bs1.vertices[1] = tmp;
    tmp[0] = 2;
    tmp[1] = 0;
    bs1.vertices[2] = tmp;

    ConvexPolyhedron<LocalVector> bs2;
    bs2.vertices.resize(3);
    tmp[0] = 4;
    tmp[1] = 2;
    bs2.vertices[0] = tmp;
    tmp[0] = 4;
    tmp[1] = 4;
    bs2.vertices[1] = tmp;
    tmp[0] = 6;
    tmp[1] = 2;
    bs2.vertices[2] = tmp;

    test(bs1, bs2, analyticalDistance);
  }
}
